{% extends "admin/change_form.html" %}
{% load l10n i18n %}

{% block submit_buttons_bottom %}
    {{ block.super }}
    <div class="submit-row">
      <input type="submit" value="{% trans 'Duplicate' %}" name="_duplicate">
    </div>
{% endblock %}
